@@ -1,3 +1,5 @@
+# noqa: D100
+
 import logging
 from typing import Any, Dict, List, Optional, Set, Tuple, Union
 
@@ -36,10 +38,10 @@ def pop_max_expr(
     pops_to_exclude: Optional[Set[str]] = None,
 ) -> hl.expr.StructExpression:
     """
-    Creates an expression containing popmax: the frequency information about the population
-    that has the highest AF from the populations provided in `freq_meta`,
-    excluding those specified in `pops_to_exclude`.
-    Only frequencies from adj populations are considered.
+
+    Create an expression containing the frequency information about the population that has the highest AF in `freq_meta`.
+
+    Populations specified in `pops_to_exclude` are excluded and only frequencies from adj populations are considered.
 
     This resulting struct contains the following fields:
 
@@ -55,7 +57,13 @@ def pop_max_expr(
 
     :return: Popmax struct
     """
-    _pops_to_exclude = hl.literal(pops_to_exclude)
+    _pops_to_exclude = (
+        hl.literal(pops_to_exclude)
+        if pops_to_exclude is not None
+        else hl.empty_set(hl.tstr)
+    )
+
+    # pylint: disable=invalid-unary-operand-type
     popmax_freq_indices = hl.range(0, hl.len(freq_meta)).filter(
         lambda i: (hl.set(freq_meta[i].keys()) == {"group", "pop"})
         & (freq_meta[i]["group"] == "adj")
@@ -76,8 +84,9 @@ def project_max_expr(
     n_projects: int = 5,
 ) -> hl.expr.ArrayExpression:
     """
-    Creates an expression that computes allele frequency information by project for the `n_projects` with the largest AF at this row.
-    This return an array with one element per non-reference allele.
+    Create an expression that computes allele frequency information by project for the `n_projects` with the largest AF at this row.
+
+    Will return an array with one element per non-reference allele.
 
     Each of these elements is itself an array of structs with the following fields:
 
@@ -98,7 +107,6 @@ def project_max_expr(
     :param n_projects: Maximum number of projects to return for each row
     :return: projectmax expression
     """
-
     n_alleles = hl.len(alleles_expr)
 
     # compute call stats by  project
@@ -140,7 +148,8 @@ def faf_expr(
     faf_thresholds: List[float] = [0.95, 0.99],
 ) -> Tuple[hl.expr.ArrayExpression, List[Dict[str, str]]]:
     """
-    Calculates the filtering allele frequency (FAF) for each threshold specified in `faf_thresholds`.
+    Calculate the filtering allele frequency (FAF) for each threshold specified in `faf_thresholds`.
+
     See http://cardiodb.org/allelefrequencyapp/ for more information.
 
     The FAF is computed for each of the following population stratification if found in `freq_meta`:
@@ -164,8 +173,12 @@ def faf_expr(
     :return: (FAF expression, FAF metadata)
     """
     _pops_to_exclude = (
-        hl.literal(pops_to_exclude) if pops_to_exclude is not None else {}
+        hl.literal(pops_to_exclude)
+        if pops_to_exclude is not None
+        else hl.empty_set(hl.tstr)
     )
+
+    # pylint: disable=invalid-unary-operand-type
     faf_freq_indices = hl.range(0, hl.len(freq_meta)).filter(
         lambda i: (freq_meta[i].get("group") == "adj")
         & (
@@ -227,7 +240,7 @@ def qual_hist_expr(
     adj_expr: Optional[hl.expr.BooleanExpression] = None,
 ) -> hl.expr.StructExpression:
     """
-    Returns a struct expression with genotype quality histograms based on the arguments given (dp, gq, ad).
+    Return a struct expression with genotype quality histograms based on the arguments given (dp, gq, ad).
 
     .. note::
 
@@ -292,7 +305,7 @@ def age_hists_expr(
     n_bins: int = 10,
 ) -> hl.expr.StructExpression:
     """
-    Returns a StructExpression with the age histograms for hets and homs.
+    Return a StructExpression with the age histograms for hets and homs.
 
     :param adj_expr: Entry expression containing whether a genotype is high quality (adj) or not
     :param gt_expr: Entry expression containing the genotype
@@ -323,8 +336,12 @@ def annotate_freq(
     downsamplings: Optional[List[int]] = None,
 ) -> hl.MatrixTable:
     """
-    Adds a row annotation `freq` to the input `mt` with stratified allele frequencies,
-    and a global annotation `freq_meta` with metadata.
+    Annotate `mt` with stratified allele frequencies.
+
+    The output Matrix table will include:
+        - row annotation `freq` containing the stratified allele frequencies
+        - global annotation `freq_meta` with metadata
+        - global annotation `freq_sample_count` with sample count information
 
     .. note::
 
@@ -352,6 +369,11 @@ def annotate_freq(
     Each element of the list contains metadata on a frequency stratification and the index in the list corresponds
     to the index of that frequency stratification in the `freq` row annotation.
 
+    .. rubric:: Global `freq_sample_count` annotation
+
+    The global annotation `freq_sample_count` is added to the input `mt`. This is a sample count per sample grouping
+    defined in the `freq_meta` global annotation.
+
     .. rubric:: The `downsamplings` parameter
 
     If the `downsamplings` parameter is used, frequencies will be computed for all samples and by population
@@ -368,7 +390,6 @@ def annotate_freq(
     :param downsamplings: When specified, frequencies are computed by downsampling the data to the number of samples given in the list. Note that if `pop_expr` is specified, downsamplings by population is also computed.
     :return: MatrixTable with `freq` annotation
     """
-
     if subpop_expr is not None and pop_expr is None:
         raise NotImplementedError(
             "annotate_freq requires pop_expr when using subpop_expr"
@@ -417,7 +438,7 @@ def annotate_freq(
             downsamplings = sorted(
                 [x for x in downsamplings if x <= sum(cut_data.get("pop").values())]
             )
-        logger.info(f"Found {len(downsamplings)} downsamplings: {downsamplings}")
+        logger.info("Found %d downsamplings: %s", len(downsamplings), downsamplings)
 
         # Shuffle the samples, then create a global index for downsampling
         # And a pop-index if pops were provided
@@ -486,22 +507,33 @@ def annotate_freq(
         + sample_group_filters
     )
 
+    freq_sample_count = mt.aggregate_cols(
+        [hl.agg.count_where(x[1]) for x in sample_group_filters]
+    )
+
     # Annotate columns with group_membership
     mt = mt.annotate_cols(group_membership=[x[1] for x in sample_group_filters])
 
-    # Create and annotate global expression with meta information
+    # Create and annotate global expression with meta and sample count information
     freq_meta_expr = [
         dict(**sample_group[0], group="adj") for sample_group in sample_group_filters
     ]
     freq_meta_expr.insert(1, {"group": "raw"})
-    mt = mt.annotate_globals(freq_meta=freq_meta_expr)
+    freq_sample_count.insert(1, freq_sample_count[0])
+    mt = mt.annotate_globals(
+        freq_meta=freq_meta_expr, freq_sample_count=freq_sample_count,
+    )
 
     # Create frequency expression array from the sample groups
+    # Adding sample_group_filters_range_array to reduce memory usage in this array_agg
+    mt = mt.annotate_rows(
+        sample_group_filters_range_array=hl.range(len(sample_group_filters))
+    )
     freq_expr = hl.agg.array_agg(
         lambda i: hl.agg.filter(
             mt.group_membership[i] & mt.adj, hl.agg.call_stats(mt.GT, mt.alleles)
         ),
-        hl.range(len(sample_group_filters)),
+        mt.sample_group_filters_range_array,
     )
 
     # Insert raw as the second element of the array
@@ -535,7 +567,7 @@ def get_lowqual_expr(
     indel_phred_het_prior: int = 39,  # 1/8,000
 ) -> Union[hl.expr.BooleanExpression, hl.expr.ArrayExpression]:
     """
-    Computes lowqual threshold expression for either split or unsplit alleles based on QUALapprox or AS_QUALapprox
+    Compute lowqual threshold expression for either split or unsplit alleles based on QUALapprox or AS_QUALapprox.
 
     .. note::
 
@@ -551,7 +583,6 @@ def get_lowqual_expr(
     :param indel_phred_het_prior: Phred-scaled indel heterozygosity prior (30 = 1/1000 bases, GATK default)
     :return: lowqual expression (BooleanExpression if `qual_approx_expr`is Numeric, Array[BooleanExpression] if `qual_approx_expr` is ArrayNumeric)
     """
-
     min_snv_qual = snv_phred_threshold + snv_phred_het_prior
     min_indel_qual = indel_phred_threshold + indel_phred_het_prior
     min_mixed_qual = max(min_snv_qual, min_indel_qual)
@@ -589,7 +620,8 @@ def get_annotations_hists(
     log10_annotations: List[str] = ["DP"],
 ) -> Dict[str, hl.expr.StructExpression]:
     """
-    Creates histograms for variant metrics in ht.info.
+    Create histograms for variant metrics in ht.info.
+
     Used when creating site quality distribution json files.
 
     :param ht: Table with variant metrics
@@ -615,16 +647,16 @@ def create_frequency_bins_expr(
     AC: hl.expr.NumericExpression, AF: hl.expr.NumericExpression
 ) -> hl.expr.StringExpression:
     """
-    Creates bins for frequencies in preparation for aggregating QUAL by frequency bin.
+    Create bins for frequencies in preparation for aggregating QUAL by frequency bin.
 
-    Bins: 
+    Bins:
         - singleton
-        - doubleton 
-        - 0.00005 
-        - 0.0001 
+        - doubleton
+        - 0.00005
+        - 0.0001
         - 0.0002
         - 0.0005
-        - 0.001, 
+        - 0.001,
         - 0.002
         - 0.005
         - 0.01
@@ -634,7 +666,7 @@ def create_frequency_bins_expr(
         - 0.2
         - 0.5
         - 1
-    
+
     NOTE: Frequencies should be frequencies from raw data.
     Used when creating site quality distribution json files.
 
@@ -674,10 +706,11 @@ def get_adj_expr(
     adj_gq: int = 20,
     adj_dp: int = 10,
     adj_ab: float = 0.2,
-    haploid_adj_dp: int = 10,
+    haploid_adj_dp: int = 5,
 ) -> hl.expr.BooleanExpression:
     """
-    Gets adj genotype annotation.
+    Get adj genotype annotation.
+
     Defaults correspond to gnomAD values.
     """
     return (
@@ -700,10 +733,11 @@ def annotate_adj(
     adj_gq: int = 20,
     adj_dp: int = 10,
     adj_ab: float = 0.2,
-    haploid_adj_dp: int = 10,
+    haploid_adj_dp: int = 5,
 ) -> hl.MatrixTable:
     """
-    Annotate genotypes with adj criteria (assumes diploid)
+    Annotate genotypes with adj criteria (assumes diploid).
+
     Defaults correspond to gnomAD values.
     """
     return mt.annotate_entries(
@@ -714,9 +748,7 @@ def annotate_adj(
 
 
 def add_variant_type(alt_alleles: hl.expr.ArrayExpression) -> hl.expr.StructExpression:
-    """
-    Get Struct of variant_type and n_alt_alleles from ArrayExpression of Strings (all alleles)
-    """
+    """Get Struct of variant_type and n_alt_alleles from ArrayExpression of Strings (all alleles)."""
     ref = alt_alleles[0]
     alts = alt_alleles[1:]
     non_star_alleles = hl.filter(lambda a: a != "*", alts)
@@ -736,33 +768,32 @@ def add_variant_type(alt_alleles: hl.expr.ArrayExpression) -> hl.expr.StructExpr
 
 def annotation_type_is_numeric(t: Any) -> bool:
     """
-    Given an annotation type, returns whether it is a numerical type or not.
+    Given an annotation type, return whether it is a numerical type or not.
 
     :param t: Type to test
     :return: If the input type is numeric
     """
-    return (
-        isinstance(t, hl.tint32)
-        or isinstance(t, hl.tint64)
-        or isinstance(t, hl.tfloat32)
-        or isinstance(t, hl.tfloat64)
-    )
+    return t in (hl.tint32, hl.tint64, hl.tfloat32, hl.tfloat64)
 
 
 def annotation_type_in_vcf_info(t: Any) -> bool:
     """
     Given an annotation type, returns whether that type can be natively exported to a VCF INFO field.
-    Note types that aren't natively exportable to VCF will be converted to String on export.
+
+    .. note::
+
+        Types that aren't natively exportable to VCF will be converted to String on export.
 
     :param t: Type to test
     :return: If the input type can be exported to VCF
     """
     return (
         annotation_type_is_numeric(t)
-        or isinstance(t, hl.tstr)
-        or isinstance(t, hl.tarray)
-        or isinstance(t, hl.tset)
-        or isinstance(t, hl.tbool)
+        or t in (hl.tstr, hl.tbool)
+        or (
+            isinstance(t, (hl.tarray, hl.tset))
+            and annotation_type_in_vcf_info(t.element_type)
+        )
     )
 
 
@@ -802,7 +833,8 @@ def fs_from_sb(
     min_p_value: float = 1e-320,
 ) -> hl.expr.Int64Expression:
     """
-    Computes `FS` (Fisher strand balance) annotation from  the `SB` (strand balance table) field.
+    Compute `FS` (Fisher strand balance) annotation from  the `SB` (strand balance table) field.
+
     `FS` is the phred-scaled value of the double-sided Fisher exact test on strand balance.
 
     Using default values will have the same behavior as the GATK implementation, that is:
@@ -816,8 +848,8 @@ def fs_from_sb(
     .. note::
 
         This function can either take
-        - an array of length containing the table counts: [ref fwd, ref rev, alt fwd, alt rev]
-        - an array containig 2 arrays of length 2, containing the counts: [[ref fwd, ref rev], [alt fwd, alt rev]]
+        - an array of length four containing the forward and reverse strands' counts of ref and alt alleles: [ref fwd, ref rev, alt fwd, alt rev]
+        - a two dimensional array with arrays of length two, containing the counts: [[ref fwd, ref rev], [alt fwd, alt rev]]
 
     GATK code here: https://github.com/broadinstitute/gatk/blob/master/src/main/java/org/broadinstitute/hellbender/tools/walkers/annotator/FisherStrand.java
 
@@ -869,10 +901,45 @@ def fs_from_sb(
     )
 
 
+def sor_from_sb(
+    sb: Union[hl.expr.ArrayNumericExpression, hl.expr.ArrayExpression]
+) -> hl.expr.Float64Expression:
+    """
+    Compute `SOR` (Symmetric Odds Ratio test) annotation from  the `SB` (strand balance table) field.
+
+    .. note::
+
+        This function can either take
+        - an array of length four containing the forward and reverse strands' counts of ref and alt alleles: [ref fwd, ref rev, alt fwd, alt rev]
+        - a two dimensional array with arrays of length two, containing the counts: [[ref fwd, ref rev], [alt fwd, alt rev]]
+
+    GATK code here: https://github.com/broadinstitute/gatk/blob/master/src/main/java/org/broadinstitute/hellbender/tools/walkers/annotator/StrandOddsRatio.java
+
+    :param sb: Count of ref/alt reads on each strand
+    :return: SOR value
+    """
+    if not isinstance(sb, hl.expr.ArrayNumericExpression):
+        sb = hl.bind(lambda x: hl.flatten(x), sb)
+
+    sb = sb.map(lambda x: hl.float64(x) + 1)
+
+    ref_fw = sb[0]
+    ref_rv = sb[1]
+    alt_fw = sb[2]
+    alt_rv = sb[3]
+    symmetrical_ratio = ((ref_fw * alt_rv) / (alt_fw * ref_rv)) + (
+        (alt_fw * ref_rv) / (ref_fw * alt_rv)
+    )
+    ref_ratio = hl.min(ref_rv, ref_fw) / hl.max(ref_rv, ref_fw)
+    alt_ratio = hl.min(alt_fw, alt_rv) / hl.max(alt_fw, alt_rv)
+    sor = hl.log(symmetrical_ratio) + hl.log(ref_ratio) - hl.log(alt_ratio)
+
+    return sor
+
+
 def bi_allelic_expr(t: Union[hl.Table, hl.MatrixTable]) -> hl.expr.BooleanExpression:
     """
-    Returns a boolean expression selecting bi-allelic sites only,
-    accounting for whether the input MT/HT was split.
+    Return a boolean expression selecting bi-allelic sites only, accounting for whether the input MT/HT was split.
 
     :param t: Input HT/MT
     :return: Boolean expression selecting only bi-allelic sites
@@ -882,7 +949,7 @@ def bi_allelic_expr(t: Union[hl.Table, hl.MatrixTable]) -> hl.expr.BooleanExpres
 
 def unphase_call_expr(call_expr: hl.expr.CallExpression) -> hl.expr.CallExpression:
     """
-    Generate unphased version of a call expression (which can be phased or not)
+    Generate unphased version of a call expression (which can be phased or not).
 
     :param call_expr: Input call expression
     :return: unphased call expression
@@ -893,3 +960,103 @@ def unphase_call_expr(call_expr: hl.expr.CallExpression) -> hl.expr.CallExpressi
         .when(call_expr.is_haploid(), hl.call(call_expr[0], phased=False))
         .default(hl.null(hl.tcall))
     )
+
+
+def region_flag_expr(
+    t: Union[hl.Table, hl.MatrixTable],
+    non_par: bool = True,
+    prob_regions: Dict[str, hl.Table] = None,
+) -> hl.expr.StructExpression:
+    """
+    Create a `region_flag` struct that contains flags for problematic regions (i.e., LCR, decoy, segdup, and nonpar regions).
+
+    .. note:: No hg38 resources for decoy or self chain are available yet.
+
+    :param t: Input Table/MatrixTable
+    :param non_par: If True, flag loci that occur within pseudoautosomal regions on sex chromosomes
+    :param prob_regions: If supplied, flag loci that occur within regions defined in Hail Table(s)
+    :return: `region_flag` struct row annotation
+    """
+    prob_flags_expr = (
+        {"non_par": (t.locus.in_x_nonpar() | t.locus.in_y_nonpar())} if non_par else {}
+    )
+
+    if prob_regions is not None:
+        prob_flags_expr.update(
+            {
+                region_name: hl.is_defined(region_table[t.locus])
+                for region_name, region_table in prob_regions.items()
+            }
+        )
+
+    return hl.struct(**prob_flags_expr)
+
+
+def missing_callstats_expr() -> hl.expr.StructExpression:
+    """
+    Create a missing callstats struct for insertion into frequency annotation arrays when data is missing.
+
+    :return: Hail Struct with missing values for each callstats element
+    """
+    return hl.struct(
+        AC=hl.missing(hl.tint32),
+        AF=hl.missing(hl.tfloat64),
+        AN=hl.missing(hl.tint32),
+        homozygote_count=hl.missing(hl.tint32),
+    )
+
+
+def set_female_y_metrics_to_na_expr(
+    t: Union[hl.Table, hl.MatrixTable]
+) -> hl.expr.ArrayExpression:
+    """
+    Set Y-variant frequency callstats for female-specific metrics to missing structs.
+
+    .. note:: Requires freq, freq_meta, and freq_index_dict annotations to be present in Table or MatrixTable
+
+    :param t: Table or MatrixTable for which to adjust female metrics
+    :return: Hail array expression to set female Y-variant metrics to missing values
+    """
+    female_idx = hl.map(
+        lambda x: t.freq_index_dict[x],
+        hl.filter(lambda x: x.contains("XX"), t.freq_index_dict.keys()),
+    )
+    freq_idx_range = hl.range(hl.len(t.freq_meta))
+
+    new_freq_expr = hl.if_else(
+        (t.locus.in_y_nonpar() | t.locus.in_y_par()),
+        hl.map(
+            lambda x: hl.if_else(
+                female_idx.contains(x), missing_callstats_expr(), t.freq[x]
+            ),
+            freq_idx_range,
+        ),
+        t.freq,
+    )
+
+    return new_freq_expr
+
+
+def hemi_expr(
+    locus: hl.expr.LocusExpression,
+    sex_expr: hl.expr.StringExpression,
+    gt: hl.expr.CallExpression,
+    male_str: str = "XY",
+) -> hl.expr.BooleanExpression:
+    """
+    Return whether genotypes are hemizygous.
+
+    Return missing expression if locus is not in chrX/chrY non-PAR regions.
+
+    :param locus: Input locus.
+    :param sex_expr: Input StringExpression indicating whether sample is XX or XY.
+    :param gt: Input genotype.
+    :param xy_str: String indicating whether sample is XY. Default is "XY".
+    :return: BooleanExpression indicating whether genotypes are hemizygous.
+    """
+    return hl.or_missing(
+        locus.in_x_nonpar() | locus.in_y_nonpar(),
+        # Haploid genotypes have a single integer, so checking if
+        # mt.GT[0] is alternate allele
+        gt.is_haploid() & (sex_expr == male_str) & (gt[0] == 1),
+    )
