@@ -1,8 +1,15 @@
+# noqa: D100
+
 import logging
 from typing import Dict, List, Optional, Union
 
 import hail as hl
-from gnomad.utils.annotations import fs_from_sb, get_adj_expr, get_lowqual_expr
+from gnomad.utils.annotations import (
+    fs_from_sb,
+    get_adj_expr,
+    get_lowqual_expr,
+    sor_from_sb,
+)
 from gnomad.utils.intervals import interval_length, union_intervals
 from gnomad.utils.reference_genome import get_reference_genome
 
@@ -21,8 +28,7 @@ INFO_ARRAY_SUM_AGG_FIELDS = ["SB", "RAW_MQandDP"]
 
 def compute_last_ref_block_end(mt: hl.MatrixTable) -> hl.Table:
     """
-    This function takes a sparse MT and computes for each row the genomic position of the
-    most upstream reference block overlapping that row.
+    Compute the genomic position of the most upstream reference block overlapping each row on a sparse MT.
 
     Note that since reference blocks do not extend beyond contig boundaries, only the position is kept.
 
@@ -72,7 +78,7 @@ def compute_last_ref_block_end(mt: hl.MatrixTable) -> hl.Table:
             ht.locus.position,
         )
     )
-    return ht.select_globals()
+    return ht.select_globals().key_by("locus")
 
 
 def densify_sites(
@@ -82,7 +88,7 @@ def densify_sites(
     semi_join_rows: bool = True,
 ) -> hl.MatrixTable:
     """
-    Creates a dense version of the input sparse MT at the sites in `sites_ht` reading the minimal amount of data required.
+    Create a dense version of the input sparse MT at the sites in `sites_ht` reading the minimal amount of data required.
 
     Note that only rows that appear both in `mt` and `sites_ht` are returned.
 
@@ -147,16 +153,16 @@ def _get_info_agg_expr(
     prefix: str = "",
 ) -> Dict[str, hl.expr.Aggregation]:
     """
-    Helper function containing code to create Aggregators for both site or AS info expression aggregations.
+    Create Aggregators for both site or AS info expression aggregations.
 
-    Notes:
+    .. note::
 
-    1. If `SB` is specified in array_sum_agg_fields, it will be aggregated as `AS_SB_TABLE`, according to GATK standard nomenclature.
-    2. If `RAW_MQandDP` is specified in array_sum_agg_fields, it will be used for the `MQ` calculation and then dropped according to GATK recommendation.
-    3. If `RAW_MQ` and `MQ_DP` are given, they will be used for the `MQ` calculation and then dropped according to GATK recommendation.
-    4. If the fields to be aggregate (`sum_agg_fields`, `int32_sum_agg_fields`, `median_agg_fields`) are passed as list of str,
-       then they should correspond to entry fields in `mt` or in `mt.gvcf_info`.
-       Priority is given to entry fields in `mt` over those in `mt.gvcf_info` in case of a name clash.
+        - If `SB` is specified in array_sum_agg_fields, it will be aggregated as `AS_SB_TABLE`, according to GATK standard nomenclature.
+        - If `RAW_MQandDP` is specified in array_sum_agg_fields, it will be used for the `MQ` calculation and then dropped according to GATK recommendation.
+        - If `RAW_MQ` and `MQ_DP` are given, they will be used for the `MQ` calculation and then dropped according to GATK recommendation.
+        - If the fields to be aggregate (`sum_agg_fields`, `int32_sum_agg_fields`, `median_agg_fields`) are passed as
+          list of str, then they should correspond to entry fields in `mt` or in `mt.gvcf_info`.
+        - Priority is given to entry fields in `mt` over those in `mt.gvcf_info` in case of a name clash.
 
     :param mt: Input MT
     :param sum_agg_fields: Fields to aggregate using sum.
@@ -232,16 +238,18 @@ def _get_info_agg_expr(
     mq_tuple = None
     if f"{prefix}RAW_MQandDP" in agg_expr:
         logger.info(
-            f"Computing {prefix}MQ as sqrt({prefix}RAW_MQandDP[0]/{prefix}RAW_MQandDP[1]). "
-            f"Note that {prefix}MQ will be set to 0 if {prefix}RAW_MQandDP[1] == 0."
+            "Computing %sMQ as sqrt(%sRAW_MQandDP[0]/%sRAW_MQandDP[1]). "
+            "Note that %sMQ will be set to 0 if %sRAW_MQandDP[1] == 0.",
+            *[prefix] * 5,
         )
         mq_tuple = agg_expr.pop(f"{prefix}RAW_MQandDP")
     elif f"{prefix}RAW_MQ" in agg_expr and f"{prefix}MQ_DP" in agg_expr:
         logger.info(
-            f"Computing {prefix}MQ as sqrt({prefix}MQ_DP/{prefix}RAW_MQ). "
-            f"Note that MQ will be set to 0 if {prefix}RAW_MQ == 0."
+            "Computing %sMQ as sqrt(%sRAW_MQ/%sMQ_DP). "
+            "Note that MQ will be set to 0 if %sRAW_MQ == 0.",
+            *[prefix] * 4,
         )
-        mq_tuple = (agg_expr.pop(f"{prefix}MQ_DP"), agg_expr.pop(f"{prefix}RAW_MQ"))
+        mq_tuple = (agg_expr.pop(f"{prefix}RAW_MQ"), agg_expr.pop(f"{prefix}MQ_DP"))
 
     if mq_tuple is not None:
         agg_expr[f"{prefix}MQ"] = hl.cond(
@@ -251,8 +259,9 @@ def _get_info_agg_expr(
     # If both VarDP and QUALapprox are present, also compute QD.
     if f"{prefix}VarDP" in agg_expr and f"{prefix}QUALapprox" in agg_expr:
         logger.info(
-            f"Computing {prefix}QD as {prefix}QUALapprox/{prefix}VarDP. "
-            f"Note that {prefix}QD will be set to 0 if {prefix}VarDP == 0."
+            "Computing %sQD as %sQUALapprox/%sVarDP. "
+            "Note that %sQD will be set to 0 if %sVarDP == 0.",
+            *[prefix] * 5,
         )
         var_dp = hl.int32(hl.agg.sum(int32_sum_agg_fields["VarDP"]))
         agg_expr[f"{prefix}QD"] = hl.cond(
@@ -280,23 +289,26 @@ def get_as_info_expr(
     array_sum_agg_fields: Union[
         List[str], Dict[str, hl.expr.ArrayNumericExpression]
     ] = INFO_ARRAY_SUM_AGG_FIELDS,
+    alt_alleles_range_array_field: str = "alt_alleles_range_array",
 ) -> hl.expr.StructExpression:
     """
-    Returns an allele-specific annotation Struct containing typical VCF INFO fields from GVCF INFO fields stored in the MT entries.
+    Return an allele-specific annotation Struct containing typical VCF INFO fields from GVCF INFO fields stored in the MT entries.
 
-    Notes:
+    .. note::
 
-    1. If `SB` is specified in array_sum_agg_fields, it will be aggregated as `AS_SB_TABLE`, according to GATK standard nomenclature.
-    2. If `RAW_MQandDP` is specified in array_sum_agg_fields, it will be used for the `MQ` calculation and then dropped according to GATK recommendation.
-    3. If `RAW_MQ` and `MQ_DP` are given, they will be used for the `MQ` calculation and then dropped according to GATK recommendation.
-    4. If the fields to be aggregate (`sum_agg_fields`, `int32_sum_agg_fields`, `median_agg_fields`) are passed as list of str,
-       then they should correspond to entry fields in `mt` or in `mt.gvcf_info`.
-       Priority is given to entry fields in `mt` over those in `mt.gvcf_info` in case of a name clash.
+        - If `SB` is specified in array_sum_agg_fields, it will be aggregated as `AS_SB_TABLE`, according to GATK standard nomenclature.
+        - If `RAW_MQandDP` is specified in array_sum_agg_fields, it will be used for the `MQ` calculation and then dropped according to GATK recommendation.
+        - If `RAW_MQ` and `MQ_DP` are given, they will be used for the `MQ` calculation and then dropped according to GATK recommendation.
+        - If the fields to be aggregate (`sum_agg_fields`, `int32_sum_agg_fields`, `median_agg_fields`) are passed as list of str,
+          then they should correspond to entry fields in `mt` or in `mt.gvcf_info`.
+        - Priority is given to entry fields in `mt` over those in `mt.gvcf_info` in case of a name clash.
 
     :param mt: Input Matrix Table
     :param sum_agg_fields: Fields to aggregate using sum.
     :param int32_sum_agg_fields: Fields to aggregate using sum using int32.
     :param median_agg_fields: Fields to aggregate using (approximate) median.
+    :param array_sum_agg_fields: Fields to aggregate using array sum.
+    :param alt_alleles_range_array_field: Annotation containing an array of the range of alternate alleles e.g., `hl.range(1, hl.len(mt.alleles))`
     :return: Expression containing the AS info fields
     """
     if "DP" in list(sum_agg_fields) + list(int32_sum_agg_fields):
@@ -319,11 +331,18 @@ def get_as_info_expr(
     if "AS_SB" in agg_expr:
         agg_expr["AS_SB_TABLE"] = agg_expr.pop("AS_SB")
 
+    if alt_alleles_range_array_field not in mt.row or mt[
+        alt_alleles_range_array_field
+    ].dtype != hl.dtype("array<int32>"):
+        msg = f"'get_as_info_expr' expected a row field '{alt_alleles_range_array_field}' of type array<int32>"
+        logger.error(msg)
+        raise ValueError(msg)
+
     # Modify aggregations to aggregate per allele
     agg_expr = {
         f: hl.agg.array_agg(
             lambda ai: hl.agg.filter(mt.LA.contains(ai), expr),
-            hl.range(1, hl.len(mt.alleles)),
+            mt[alt_alleles_range_array_field],
         )
         for f, expr in agg_expr.items()
     }
@@ -347,6 +366,9 @@ def get_as_info_expr(
             AS_FS=hl.range(1, hl.len(mt.alleles)).map(
                 lambda i: fs_from_sb(as_sb_table[0].extend(as_sb_table[i]))
             ),
+            AS_SOR=hl.range(1, hl.len(mt.alleles)).map(
+                lambda i: sor_from_sb(as_sb_table[0].extend(as_sb_table[i]))
+            ),
         )
 
     return info
@@ -368,15 +390,15 @@ def get_site_info_expr(
     ] = INFO_ARRAY_SUM_AGG_FIELDS,
 ) -> hl.expr.StructExpression:
     """
-    Creates a site-level annotation Struct aggregating typical VCF INFO fields from GVCF INFO fields stored in the MT entries.
+    Create a site-level annotation Struct aggregating typical VCF INFO fields from GVCF INFO fields stored in the MT entries.
 
-    Notes:
+    .. note::
 
-    1. If `RAW_MQandDP` is specified in array_sum_agg_fields, it will be used for the `MQ` calculation and then dropped according to GATK recommendation.
-    2. If `RAW_MQ` and `MQ_DP` are given, they will be used for the `MQ` calculation and then dropped according to GATK recommendation.
-    3. If the fields to be aggregate (`sum_agg_fields`, `int32_sum_agg_fields`, `median_agg_fields`) are passed as list of str,
-       then they should correspond to entry fields in `mt` or in `mt.gvcf_info`.
-       Priority is given to entry fields in `mt` over those in `mt.gvcf_info` in case of a name clash.
+        - If `RAW_MQandDP` is specified in array_sum_agg_fields, it will be used for the `MQ` calculation and then dropped according to GATK recommendation.
+        - If `RAW_MQ` and `MQ_DP` are given, they will be used for the `MQ` calculation and then dropped according to GATK recommendation.
+        - If the fields to be aggregate (`sum_agg_fields`, `int32_sum_agg_fields`, `median_agg_fields`) are passed as
+          list of str, then they should correspond to entry fields in `mt` or in `mt.gvcf_info`.
+        - Priority is given to entry fields in `mt` over those in `mt.gvcf_info` in case of a name clash.
 
     :param mt: Input Matrix Table
     :param sum_agg_fields: Fields to aggregate using sum.
@@ -397,9 +419,11 @@ def get_site_info_expr(
         array_sum_agg_fields=array_sum_agg_fields,
     )
 
-    # Add FS if SB is present
+    # Add FS and SOR if SB is present
     # This is done outside of _get_info_agg_expr as the behavior is different in site vs allele-specific versions
-    agg_expr["FS"] = fs_from_sb(agg_expr["SB"])
+    if "SB" in agg_expr:
+        agg_expr["FS"] = fs_from_sb(agg_expr["SB"])
+        agg_expr["SOR"] = sor_from_sb(agg_expr["SB"])
 
     # Run aggregator on non-ref genotypes
     info = hl.agg.filter(
@@ -418,9 +442,11 @@ def default_compute_info(
     mt: hl.MatrixTable, site_annotations: bool = False, n_partitions: int = 5000
 ) -> hl.Table:
     """
-    Computes a HT with the typical GATK allele-specific (AS) info fields 
-    as well as ACs and lowqual fields.
-    Note that this table doesn't split multi-allelic sites.
+    Compute a HT with the typical GATK allele-specific (AS) info fields as well as ACs and lowqual fields.
+
+    .. note::
+
+        This table doesn't split multi-allelic sites.
 
     :param mt: Input MatrixTable. Note that this table should be filtered to nonref sites.
     :param site_annotations: Whether to also generate site level info fields. Default is False.
@@ -431,6 +457,9 @@ def default_compute_info(
     # Move gvcf info entries out from nested struct
     mt = mt.transmute_entries(**mt.gvcf_info)
 
+    # Adding alt_alleles_range_array as a required annotation for get_as_info_expr to reduce memory usage
+    mt = mt.annotate_rows(alt_alleles_range_array=hl.range(1, hl.len(mt.alleles)))
+
     # Compute AS info expr
     info_expr = get_as_info_expr(mt)
 
@@ -482,7 +511,7 @@ def split_info_annotation(
     info_expr: hl.expr.StructExpression, a_index: hl.expr.Int32Expression
 ) -> hl.expr.StructExpression:
     """
-    Splits multi-allelic allele-specific info fields.
+    Split multi-allelic allele-specific info fields.
 
     :param info_expr: Field containing info struct.
     :param a_index: Allele index. Output by hl.split_multi or hl.split_multi_hts.
@@ -504,7 +533,7 @@ def split_lowqual_annotation(
     lowqual_expr: hl.expr.ArrayExpression, a_index: hl.expr.Int32Expression
 ) -> hl.expr.BooleanExpression:
     """
-    Splits multi-allelic low QUAL annotation.
+    Split multi-allelic low QUAL annotation.
 
     :param lowqual_expr: Field containing low QUAL annotation.
     :param a_index: Allele index. Output by hl.split_multi or hl.split_multi_hts.
@@ -515,26 +544,31 @@ def split_lowqual_annotation(
 
 def impute_sex_ploidy(
     mt: hl.MatrixTable,
-    excluded_intervals: Optional[hl.Table] = None,
-    included_intervals: Optional[hl.Table] = None,
+    excluded_calling_intervals: Optional[hl.Table] = None,
+    included_calling_intervals: Optional[hl.Table] = None,
     normalization_contig: str = "chr20",
     chr_x: Optional[str] = None,
     chr_y: Optional[str] = None,
-) -> hl.Table:  # TODO: For exomes, calling intervals need to be added
+) -> hl.Table:
     """
-    Imputes sex ploidy from a sparse Matrix Table by normalizing the coverage of chromosomes X and Y using
-    the coverage of an autosomal chromosome (by default chr20).
-    Coverage is computed using the median block coverage (summed over the block size) and the non-ref coverage at non-ref genotypes.
+    Impute sex ploidy from a sparse MatrixTable.
+
+    Sex ploidy is imputed by normalizing the coverage of chromosomes X and Y using the coverage of an autosomal
+    chromosome (by default chr20).
+
+    Coverage is computed using the median block coverage (summed over the block size) and the non-ref coverage at
+    non-ref genotypes.
 
     :param mt: Input sparse Matrix Table
-    :param excluded_intervals: Optional table of intervals to exclude from the computation.
-    :param included_intervals: Optional table of intervals to use in the computation. REQUIRED for exomes.
+    :param excluded_calling_intervals: Optional table of intervals to exclude from the computation.
+        Used only when determining contig size (not used when computing chromosome depth).
+    :param included_calling_intervals: Optional table of intervals to use in the computation.
+        Used only when determining contig size (not used when computing chromosome depth).
     :param normalization_contig: Which chromosome to normalize by
     :param chr_x: Optional X Chromosome contig name (by default uses the X contig in the reference)
     :param chr_y: Optional Y Chromosome contig name (by default uses the Y contig in the reference)
     :return: Table with mean coverage over chromosomes 20, X and Y and sex chromosomes ploidy based on normalized coverage.
     """
-
     ref = get_reference_genome(mt.locus, add_sequence=True)
     if chr_x is None:
         if len(ref.x_contigs) != 1:
@@ -554,7 +588,7 @@ def impute_sex_ploidy(
         chr_y = ref.y_contigs[0]
 
     def get_contig_size(contig: str) -> int:
-        logger.info(f"Working on {contig}")
+        logger.info("Working on %s", contig)
         contig_ht = hl.utils.range_table(
             ref.contig_length(contig),
             n_partitions=int(ref.contig_length(contig) / 500_000),
@@ -570,16 +604,16 @@ def impute_sex_ploidy(
             contig_ht = contig_ht.filter(contig_ht.locus.in_y_nonpar())
 
         contig_ht = contig_ht.key_by("locus")
-        if included_intervals is not None:
+        if included_calling_intervals is not None:
             contig_ht = contig_ht.filter(
-                hl.is_defined(included_intervals[contig_ht.key])
+                hl.is_defined(included_calling_intervals[contig_ht.key])
             )
-        if excluded_intervals is not None:
+        if excluded_calling_intervals is not None:
             contig_ht = contig_ht.filter(
-                hl.is_missing(excluded_intervals[contig_ht.key])
+                hl.is_missing(excluded_calling_intervals[contig_ht.key])
             )
         contig_size = contig_ht.count()
-        logger.info(f"Contig {contig} has {contig_size} bases for coverage.")
+        logger.info("Contig %s has %d bases for coverage.", contig, contig_size)
         return contig_size
 
     def get_chr_dp_ann(chrom: str) -> hl.Table:
@@ -591,15 +625,12 @@ def impute_sex_ploidy(
         if chrom in ref.y_contigs:
             chr_mt = chr_mt.filter_rows(chr_mt.locus.in_y_nonpar())
 
-        if included_intervals is not None:
-            chr_mt = chr_mt.filter_rows(hl.is_defined(included_intervals[chr_mt.locus]))
-
         return chr_mt.select_cols(
             **{
                 f"{chrom}_mean_dp": hl.agg.sum(
                     hl.cond(
                         chr_mt.LGT.is_hom_ref(),
-                        chr_mt.DP * (chr_mt.END - chr_mt.locus.position),
+                        chr_mt.DP * (1 + chr_mt.END - chr_mt.locus.position),
                         chr_mt.DP,
                     )
                 )
@@ -631,7 +662,9 @@ def compute_coverage_stats(
     coverage_over_x_bins: List[int] = [1, 5, 10, 15, 20, 25, 30, 50, 100],
 ) -> hl.Table:
     """
-    Computes the following coverage statistics for every base of the `reference_ht` provided:
+    Compute coverage statistics for every base of the `reference_ht` provided.
+
+    The following coverage stats are calculated:
         - mean
         - median
         - total DP
@@ -646,15 +679,14 @@ def compute_coverage_stats(
     :param coverage_over_x_bins: List of boundaries for computing samples over X
     :return: Table with per-base coverage stats
     """
-
     n_samples = mt.count_cols()
     print(f"Computing coverage stats on {n_samples} samples.")
 
     # Create an outer join with the reference Table
-    mt = mt.select_entries("END", "DP")
+    mt = mt.select_entries("END", "DP").select_cols().select_rows()
     col_key_fields = list(mt.col_key)
     t = mt._localize_entries("__entries", "__cols")
-    t = t.join(reference_ht.annotate(_in_ref=True), how="outer")
+    t = t.join(reference_ht.key_by(*mt.row_key).select(_in_ref=True), how="outer")
     t = t.annotate(
         __entries=hl.or_else(
             t.__entries,
@@ -669,6 +701,9 @@ def compute_coverage_stats(
     # Filter rows where the reference is missing
     mt = mt.filter_rows(mt._in_ref)
 
+    # Unfilter entries so that entries with no ref block overlap aren't null
+    mt = mt.unfilter_entries()
+
     # Compute coverage stats
     coverage_over_x_bins = sorted(coverage_over_x_bins)
     max_coverage_bin = coverage_over_x_bins[-1]
@@ -676,7 +711,7 @@ def compute_coverage_stats(
 
     # This expression creates a counter DP -> number of samples for DP between 0 and max_coverage_bin
     coverage_counter_expr = hl.agg.counter(
-        hl.or_else(hl.min(max_coverage_bin, mt.DP), 0)
+        hl.min(max_coverage_bin, hl.or_else(mt.DP, 0))
     )
 
     # This expression aggregates the DP counter in reverse order of the coverage_over_x_bins
@@ -697,12 +732,12 @@ def compute_coverage_stats(
             )
         )
     )
-    mean_expr = hl.agg.mean(mt.DP)
+    mean_expr = hl.agg.mean(hl.or_else(mt.DP, 0))
 
     # Annotate rows now
     return mt.select_rows(
         mean=hl.cond(hl.is_nan(mean_expr), 0, mean_expr),
-        median=hl.or_else(hl.agg.approx_median(mt.DP), 0),
+        median_approx=hl.or_else(hl.agg.approx_median(hl.or_else(mt.DP, 0)), 0),
         total_DP=hl.agg.sum(mt.DP),
         **{
             f"over_{x}": count_array_expr[i] / n_samples
@@ -714,3 +749,20 @@ def compute_coverage_stats(
             )
         },
     ).rows()
+
+
+def filter_ref_blocks(
+    t: Union[hl.MatrixTable, hl.Table]
+) -> Union[hl.MatrixTable, hl.Table]:
+    """
+    Filter ref blocks out of the Table or MatrixTable.
+
+    :param t: Input MT/HT
+    :return: MT/HT with ref blocks removed
+    """
+    if isinstance(t, hl.MatrixTable):
+        t = t.filter_rows((hl.len(t.alleles) > 1))
+    else:
+        t = t.filter((hl.len(t.alleles) > 1))
+
+    return t
